@@ -1,4 +1,4 @@
-# Copyright (C) 2006--2017  Kipp Cannon
+# Copyright (C) 2006--2019  Kipp Cannon
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -77,6 +77,9 @@ import itertools
 import math
 import numpy
 import operator
+import os
+import socket
+import time
 import warnings
 from xml import sax
 
@@ -113,11 +116,11 @@ def New(cls, columns = None, **kwargs):
 	>>> tbl = New(ProcessTable, [u"process_id", u"start_time", u"end_time", u"comment"])
 	>>> tbl.write(sys.stdout)	# doctest: +NORMALIZE_WHITESPACE
 	<Table Name="process:table">
-		<Column Type="int_8s" Name="process_id"/>
-		<Column Type="int_4s" Name="start_time"/>
-		<Column Type="int_4s" Name="end_time"/>
-		<Column Type="lstring" Name="comment"/>
-		<Stream Delimiter="," Type="Local" Name="process:table">
+		<Column Name="process_id" Type="int_8s"/>
+		<Column Name="start_time" Type="int_4s"/>
+		<Column Name="end_time" Type="int_4s"/>
+		<Column Name="comment" Type="lstring"/>
+		<Stream Name="process:table" Delimiter="," Type="Local">
 		</Stream>
 	</Table>
 	"""
@@ -282,7 +285,27 @@ class instrumentsproperty(object):
 class gpsproperty(object):
 	"""
 	Descriptor used internally to implement LIGOTimeGPS-valued
-	properties.
+	properties using pairs of integer attributes on row objects, one
+	for the integer seconds part of the GPS time and one for the
+	integer nanoseconds part.
+
+	Non-LIGOTimeGPS values are converted to LIGOTimeGPS before encoding
+	them into the integer attributes.  None is allowed as a special
+	case, which is encoded by setting both column attributes to None.
+
+	For the purpose of representing the boundaries of unbounded
+	segments (open-ended time intervals), +inf and -inf are also
+	allowed, and will be encoded into and decoded out of the integer
+	attributes.  To do so, a non-standard encoding is used that makes
+	use of denormalized GPS times, that is times whose nanosecond
+	component has a magnitude greater than 999999999.  Two such values
+	are reserved for +/- infinity.  To guard against the need for
+	additional special encodings in the future, this descriptor
+	reserves all denormalized values and will not allow calling code to
+	set GPS times to those values.  Calling code must provide
+	normalized GPS times, times with nanosecond components whose
+	magnitudes are not greater than 999999999.  When decoded, the
+	values reported are segments.PosInfinity or segments.NegInfinity.
 	"""
 	def __init__(self, s_name, ns_name):
 		self.s_name = s_name
@@ -331,6 +354,19 @@ class gpsproperty(object):
 
 
 class gpsproperty_with_gmst(gpsproperty):
+	"""
+	Variant of the gpsproperty descriptor, adding support for a third
+	"GMST" column.  When assigning a time to the GPS-valued descriptor,
+	after the pair of integer attributes are set to the encoded form of
+	the GPS time, the value is retrieved and the GMST column is set to
+	the Greenwhich mean sidereal time corresponding to that GPS time.
+	Note that the conversion to sidereal time is performed after
+	encoding the GPS time into the integer seconds and nanoseconds
+	attributes, so the sidereal time will reflect any rounding that has
+	occured as a result of that encoding.  If the GPS time is set to
+	None or +inf or -inf, the sidereal time is set to that value as
+	well.
+	"""
 	def __init__(self, s_name, ns_name, gmst_name):
 		super(gpsproperty_with_gmst, self).__init__(s_name, ns_name)
 		self.gmst_name = gmst_name
@@ -343,13 +379,28 @@ class gpsproperty_with_gmst(gpsproperty):
 			# re-retrieve the value in case it required type
 			# conversion
 			gps = self.__get__(obj)
-			setattr(obj, self.gmst_name, lal.GreenwichMeanSiderealTime(gps))
+			if not isinstance(gps, segments.infinity):
+				setattr(obj, self.gmst_name, lal.GreenwichMeanSiderealTime(gps))
+			elif gps > 0:
+				setattr(obj, self.gmst_name, float("+inf"))
+			elif gps < 0:
+				setattr(obj, self.gmst_name, float("-inf"))
+			else:
+				# this should be impossible
+				raise ValueError(gps)
 
 
 class segmentproperty(object):
 	"""
 	Descriptor used internally to expose pairs of GPS-valued properties
-	as segment-valued properties.
+	as segment-valued properties.  A segment may be set to None, which
+	is encoded by setting both GPS-valued properties to None.  Likewise
+	if both GPS-valued properties are set to None then the value
+	reported by this descriptor is None, not (None, None).
+
+	See the documentation for gpsproperty for more information on the
+	encodings it uses for special values and the limitations they
+	create.
 	"""
 	def __init__(self, start_name, stop_name):
 		self.start = start_name
@@ -412,6 +463,30 @@ class ProcessTable(table.Table):
 		"""
 		return set(row.process_id for row in self if row.program == program)
 
+	@staticmethod
+	def get_username():
+		"""
+		Utility to help retrieve a sensible value for the current
+		username.  First the environment variable LOGNAME is tried,
+		if that is not set the environment variable USERNAME is
+		tried, if that is not set the password database is
+		consulted (only on Unix systems, if the import of the pwd
+		module succeeds), finally if that fails KeyError is raised.
+		"""
+		try:
+			return os.environ["LOGNAME"]
+		except KeyError:
+			pass
+		try:
+			return os.environ["USERNAME"]
+		except KeyError:
+			pass
+		try:
+			import pwd
+			return pwd.getpwuid(os.getuid())[0]
+		except (ImportError, KeyError):
+			raise KeyError
+
 
 class Process(table.Table.RowType):
 	"""
@@ -452,6 +527,63 @@ class Process(table.Table.RowType):
 	end = gpsproperty("end_time", "end_time_ns")
 	segment = segmentproperty("start", "end")
 
+	@classmethod
+	def initialized(cls, program = None, version = None, cvs_repository = None, cvs_entry_time = None, comment = None, is_online = False, jobid = 0, domain = None, instruments = None, process_id = None):
+		"""
+		Create a new Process object and initialize its attributes
+		to sensible defaults.  If not None, program, version,
+		cvs_repository, comment, and domain should all be strings
+		or unicodes.  If cvs_entry_time is not None, it must be a
+		string or unicode in the format "YYYY-MM-DD HH:MM:SS".
+		is_online should be boolean, jobid an integer.  If not
+		None, instruments must be an iterable (set, tuple, etc.) of
+		instrument names (strings or unicodes).
+
+		In addition, .node is set to the current hostname,
+		.unix_procid is set to the current process ID, .username is
+		set to the current user's name, .start_time is set to the
+		current GPS time.
+
+		Note:  if the process_id keyword argument is None (the
+		default), then the process_id attribute is not set, it is
+		left uninitialized rather than setting it to None.  It must
+		be initialized before the row object can be written to a
+		file, and so in this way the calling code is required to
+		provide a proper value for it.
+
+		Example:
+
+		>>> process = Process.initialized()
+		"""
+		self = cls(
+			program = program,
+			version = version,
+			cvs_repository = cvs_repository,
+			cvs_entry_time = lal.UTCToGPS(time.strptime(cvs_entry_time, "%Y-%m-%d %H:%M:%S +0000")) if cvs_entry_time is not None else None,
+			comment = comment,
+			is_online = int(is_online),
+			node = socket.gethostname(),
+			unix_procid = os.getpid(),
+			start_time = lal.UTCToGPS(time.gmtime()),
+			end_time = None,
+			jobid = jobid,
+			domain = domain,
+			instruments = instruments
+		)
+		try:
+			self.username = ProcessTable.get_username()
+		except KeyError:
+			self.username = None
+		if process_id is not None:
+			self.process_id = process_id
+		return self
+
+	def set_end_time_now(self):
+		"""
+		Set .end_time to the current GPS time.
+		"""
+		self.end_time = lal.UTCToGPS(time.gmtime())
+
 
 ProcessTable.RowType = Process
 
@@ -484,7 +616,7 @@ class ProcessParamsTable(table.Table):
 
 	def append(self, row):
 		if row.type is not None and row.type not in ligolwtypes.Types:
-			raise ligolw.ElementError("unrecognized type '%s'" % row.type)
+			raise ligolw.ElementError("unrecognized type '%s' for process %d param '%s'" % (row.type, row.process_id, row.param))
 		super(ProcessParamsTable, self).append(row)
 
 
@@ -650,6 +782,26 @@ class SearchSummary(table.Table.RowType):
 	in_segment = segmentproperty("in_start", "in_end")
 	out_segment = segmentproperty("out_start", "out_end")
 
+	@classmethod
+	def initialized(cls, process, shared_object = "standalone", lalwrapper_cvs_tag = "", lal_cvs_tag = "", comment = None, ifos = None, inseg = None, outseg = None, nevents = 0, nnodes = 1):
+		"""
+		Create and return a sensibly initialized row for the
+		search_summary table.  process is an initialized row for
+		the process table.
+		"""
+		return cls(
+			process_id = process.process_id,
+			shared_object = shared_object,
+			lalwrapper_cvs_tag = lalwrapper_cvs_tag,
+			lal_cvs_tag = lal_cvs_tag,
+			comment = comment or process.comment,
+			instruments = ifos if ifos is not None else process.instruments,
+			in_segment = inseg,
+			out_segment = outseg,
+			nevents = nevents,
+			nnodes = nnodes
+		)
+
 
 SearchSummaryTable.RowType = SearchSummary
 
@@ -1262,6 +1414,16 @@ class SimInspiral(table.Table.RowType):
 	600000000
 	>>> print(round(x.end_time_gmst, 8))
 	-2238.39417156
+	>>> x.distance = 100e6
+	>>> x.ra_dec = 0., 0.
+	>>> x.inclination = 0.
+	>>> x.polarization = 0.
+	>>> x.snr_geometry_factors(("H1",))
+	{'H1': 0.6773046071543202}
+	>>> x.effective_distances(("H1",))
+	{'H1': 147644056.9630077}
+	>>> x.expected_snrs({"H1": 150e6})
+	{'H1': 8.127655285851842}
 	"""
 	__slots__ = tuple(map(table.Column.ColumnName, SimInspiralTable.validcolumns))
 
@@ -1322,8 +1484,9 @@ class SimInspiral(table.Table.RowType):
 		displacement at the equator, or 33 light-ns.  Therefore,
 		the failure to do a proper retarded time calculation here
 		results in errors as large as 33 ns.  This is insignificant
-		for burst searches, but be aware that this approximation is
-		being made if the return value is used in other contexts.
+		in present applications, but be aware that this
+		approximation is being made if the return value is used in
+		other contexts.
 		"""
 		# the offset is subtracted from the time of the injection.
 		# injections are done this way so that when the triggers
@@ -1333,6 +1496,72 @@ class SimInspiral(table.Table.RowType):
 		ra, dec = self.ra_dec
 		return t_geocent + lal.TimeDelayFromEarthCenter(lal.cached_detector_by_prefix[instrument].location, ra, dec, t_geocent)
 
+	def snr_geometry_factors(self, instruments):
+		"""
+		Compute and return a dictionary of the ratios of the
+		source's physical distance to its effective distances for
+		each of the given instruments.  The expected SNR in a
+		detector is
+
+		rho_{0} = 8 * (D_horizon / D) * snr_geometry_factor,
+
+		where D_horizon is the detector's horizon distance for this
+		waveform (computed from the detector's noise spectral
+		density), and D is the source's physical distance.  The
+		geometry factor (what this method computes) depends on the
+		direction to the source with respect to the antenna beam,
+		the inclination of the source's orbital plane, and the wave
+		frame's polarization.  The combination (D / geometry
+		factor) is called the effective distance.  See Equation
+		(4.3) of arXiv:0705.1514.
+
+		See also .effective_distances(), .expected_snrs().
+		"""
+		cos2i = math.cos(self.inclination)**2.
+		# don't rely on self.gmst to be set properly
+		gmst = lal.GreenwichMeanSiderealTime(self.time_geocent)
+		snr_geometry_factors = {}
+		for instrument in instruments:
+			fp, fc = lal.ComputeDetAMResponse(
+				lal.cached_detector_by_prefix[instrument].response,
+				self.longitude, self.latitude,
+				self.polarization,
+				gmst
+			)
+			snr_geometry_factors[instrument] = math.sqrt(fp**2. * (1. + cos2i)**2. / 4. + fc**2. * cos2i)
+		return snr_geometry_factors
+
+	def effective_distances(self, instruments):
+		"""
+		Compute and return a dictionary of the effective distances
+		for this injection for the given instruments.  The expected
+		SNR in a detector is
+
+		rho_{0} = 8 * D_horizon / D_effective
+
+		where D_effective, the effective distance, is related to
+		the physical distance, D, by geometry factors
+
+		D_effective = D / (geometry factors).
+
+		See also .snr_geometry_factors(), .expected_snrs().
+		"""
+		return {instrument: self.distance / snr_geometry_factor for instrument, snr_geometry_factor in self.snr_geometry_factors(instruments).items()}
+
+	def expected_snrs(self, horizon_distances):
+		"""
+		Compute and return a dictionary of the expected SNRs for
+		this injection in the given instruments.  horizon_distances
+		is a dictionary giving the horizon distance for each of the
+		detectors for which an expected SNR is to be computed.  The
+		expected SNR in a detector is
+
+		rho_{0} = 8 * D_horizon / D_effective.
+
+		See also .effective_distances().
+		"""
+		return {instrument: 8. * horizon_distances[instrument] / effective_distance for instrument, effective_distance in self.effective_distances(horizon_distances).items()}
+
 
 SimInspiralTable.RowType = SimInspiral
 
@@ -1693,13 +1922,8 @@ class Segment(table.Table.RowType):
 	>>> segments.segmentlist(map(Segment, x & y))	# doctest: +ELLIPSIS
 	[<ligo.lw.lsctables.Segment object at 0x...>, <ligo.lw.lsctables.Segment object at 0x...>]
 
-	This implementation uses a non-standard extension to encode
-	infinite values for boundaries:  the nanoseconds component is set
-	to 0xFFFFFFFF and the seconds component set to 0x7FFFFFFF or
-	0xFFFFFFFF to indicate positive resp. negative infinity.  For this
-	reason, "denormalized" LIGOTimeGPS objects (objects whose
-	nanoseconds fields contain values exceeding +/-999999999) are
-	disallowed for use with this class.
+	Unbounded intervals are permitted.  See gpsproperty for information
+	on the encoding scheme used internally, and its limitations.
 
 	Example:
 
